@@ -5,6 +5,7 @@ Classes for handling of uncertainties in kafe2 fits.
 import abc
 import numpy as np
 import six
+import warnings
 
 import logging
 
@@ -49,8 +50,8 @@ class CovMat(object):
         return _new
 
     def __eq__(self, other):
-        return (type(other) is CovMat and np.all(self._mat == other.mat)) \
-               or np.all(self._mat == other)
+        return np.all(self._mat == other.mat) if isinstance(other, CovMat) \
+            else np.all(self._mat == other)
 
     def __len__(self):
         return self._size
@@ -108,7 +109,9 @@ class CovMat(object):
         """
         if self._cor_mat is None:
             _sqrt_vars = np.sqrt(np.diag(self._mat))
-            self._cor_mat = self._mat / np.outer(_sqrt_vars, _sqrt_vars)
+            with warnings.catch_warnings():
+                warnings.simplefilter("ignore")
+                self._cor_mat = self._mat / np.outer(_sqrt_vars, _sqrt_vars)
         return self._cor_mat
 
     @property
@@ -185,9 +188,37 @@ class GaussianErrorBase(object):
         """Pointwise error array (relative errors)."""
 
     @property
-    @abc.abstractmethod
     def reference(self):
         """Array of reference values for the error."""
+        if callable(self._reference):
+            return self._reference()
+        else:
+            return self._reference
+
+    @reference.setter
+    def reference(self, reference):
+        if reference is None:
+            ##self._reference = np.ones_like(self._err_val)
+            self._reference = None
+        elif callable(reference):
+            self._reference = reference
+        else:
+            _ref = np.asarray(reference, dtype=float)
+            # check for zero-valued references if error is marked 'relative'
+            if self.relative and np.any(_ref == 0):
+                warnings.warn(
+                    "Relative error has a reference that contains values equal to zero: %s"
+                    % _ref
+                )
+            self._reference = _ref
+
+        # invalidate error_structures opposite declared relativity type
+        if self.relative:
+            self._cov_mat = None
+            self._err = None
+        else:
+            self._cov_mat_rel = None
+            self._err_rel = None
 
     @property
     @abc.abstractmethod
@@ -347,8 +378,8 @@ class SimpleGaussianError(GaussianErrorBase):
     @error.setter
     def error(self, err_val):
         err_val = np.array(err_val, dtype=float)
-        if np.any(err_val <= 0):
-            raise ValueError("Error values must be > 0. Received: %s" % err_val)
+        if np.any(err_val < 0):
+            raise ValueError("Error values must be >= 0. Received: %s" % err_val)
         if self.relative:
             if self.reference is None:
                 raise AttributeError(
@@ -397,8 +428,8 @@ class SimpleGaussianError(GaussianErrorBase):
     @error_rel.setter
     def error_rel(self, err_val):
         err_val = np.array(err_val, dtype=float)
-        if np.any(err_val <= 0):
-            raise ValueError("Error values must be > 0. Received: %s" % err_val)
+        if np.any(err_val < 0):
+            raise ValueError("Error values must be >= 0. Received: %s" % err_val)
         if self.relative:
             self._err = None
             self._err_rel = err_val
@@ -413,34 +444,6 @@ class SimpleGaussianError(GaussianErrorBase):
         self._cov_mat = None
         self._cov_mat_rel = None
 
-    @property
-    def reference(self):
-        return self._reference
-
-    @reference.setter
-    def reference(self, reference):
-        if reference is None:
-            ##self._reference = np.ones_like(self._err_val)
-            self._reference = None
-        else:
-            _ref = np.asarray(reference, dtype=float)
-            # check for zero-valued references if error is marked 'relative'
-            if self.relative and np.any(np.isclose(_ref, 0.0, atol=1e-5)):
-                # TODO: avoid hard-coded value
-                print("WARNING: Error reference contains zero values! Replacing them "
-                      "with default minimum '%g'" % (1e-2,))
-                _default = np.ones_like(_ref) * 1e-2
-                _ref = np.where(np.isclose(_ref, 0.0, atol=1e-5), _default, _ref)
-            self._reference = _ref
-
-        # invalidate error_structures opposite declared relativity type
-        if self.relative:
-            self._cov_mat = None
-            self._err = None
-        else:
-            self._cov_mat_rel = None
-            self._err_rel = None
-
     @property
     def cov_mat(self):
         if self._cov_mat is None:
@@ -651,33 +654,6 @@ class MatrixGaussianError(GaussianErrorBase):
             self._err_rel = np.sqrt(np.diag(self.cov_mat_rel))
         return self._err_rel
 
-    @property
-    def reference(self):
-        return self._reference
-
-    @reference.setter
-    def reference(self, reference):
-        if reference is None:
-            self._reference = None
-        else:
-            _ref = np.asarray(reference, dtype=float)
-            # check for zero-valued references if error is marked 'relative'
-            if self.relative and np.any(np.isclose(_ref, 0.0, atol=1e-5)):
-                # TODO: avoid hard-coded value
-                print("WARNING: Error reference contains zero values! Replacing them "
-                      "with default minimum '%g'" % (1e-2,))
-                _default = np.ones_like(_ref) * 1e-2
-                _ref = np.where(np.isclose(_ref, 0.0, atol=1e-5), _default, _ref)
-            self._reference = _ref
-
-        # invalidate error_structures opposite declared relativity type
-        if self.relative:
-            self._cov_mat = None
-            self._err = None
-        else:
-            self._cov_mat_rel = None
-            self._err_rel = None
-
     @property
     def error_uncor(self):
         raise AttributeError("Cannot get the uncorrelated part of a 'matrix-type' error!")
