@@ -2,7 +2,7 @@ import numpy as np
 import six
 import warnings
 
-from scipy.stats import poisson, norm
+from scipy.stats import poisson, norm, chi2
 from ..io.file import FileIOMixin
 from .format import ParameterFormatter, CostFunctionFormatter
 
@@ -19,41 +19,6 @@ __all__ = ["CostFunction",
            "CostFunctionException"]
 
 
-def _generic_chi2_nuisance(data, model,
-                           nuisance_vector=np.array([]),
-                           nuisance_cor_design_mat=None,
-                           uncor_cov_mat_inverse=None,
-                           fail_on_no_matrix=False):
-
-    data = np.asarray(data)
-    model = np.asarray(model)
-
-    if model.shape != data.shape:
-        raise CostFunctionException("'data' and 'model' must have the same shape! Got %r and %r..."
-                                    % (data.shape, model.shape))
-
-    # if an uncorrelated cov-mat is given use it
-    if uncor_cov_mat_inverse is not None:
-        _inner_sum = np.squeeze(np.asarray(nuisance_vector.dot(nuisance_cor_design_mat)))
-        _nuisance_penalties = nuisance_vector.dot(nuisance_vector)
-        _chisquare = (data - model - _inner_sum).dot(uncor_cov_mat_inverse).dot(data - model - _inner_sum)
-        if np.isnan(_chisquare):
-            return np.inf
-        return _chisquare + _nuisance_penalties
-
-    # raise if uncorrelated matrix is None and the correlated is not None
-    if nuisance_vector.all() == 0.0:
-        raise CostFunctionException('y-errors must not be fully correlated!')
-
-    if fail_on_no_matrix:
-        raise np.linalg.LinAlgError("Uncorrelated Covariance matrix is singular!")
-    # chisquare without errors
-    _chisquare = (data -model).dot(data-model)
-    if np.isnan(_chisquare):
-        return np.inf
-    return _chisquare
-
-
 class CostFunctionException(Exception):
     pass
 
@@ -75,7 +40,6 @@ class CostFunction(FileIOMixin, object):
     This class provides the basic functionality used by all :py:class:`CostFunction` objects.
     These use introspection (:py:mod:`inspect`) for determining the parameter structure of the
     cost function and to ensure the function can be used as a cost function (validation).
-
     """
 
     EXCEPTION_TYPE = CostFunctionException
@@ -88,13 +52,11 @@ class CostFunction(FileIOMixin, object):
         """
         Construct :py:class:`CostFunction` object (a wrapper for a native Python function):
 
-        :param cost_function: function handle
-        :type cost_function: typing.Callable
-        :param arg_names: the names to use for the cost function arguments. If None, detect from
-        function signature.
-        :type arg_names: typing.Iterable[str]
-        :param add_constraint_cost: If :py:obj:`True`, automatically add the cost for kafe2 constraints.
-        :type add_constraint_cost: bool
+        :param typing.Callable cost_function: function handle
+        :param typing.Iterable[str] arg_names: the names to use for the cost function arguments.
+            If None, detect from function signature.
+        :param bool add_constraint_cost: If :py:obj:`True`, automatically add the cost for kafe2
+            constraints.
         """
         self._cost_function_handle = cost_function
         _signature = signature(self._cost_function_handle)
@@ -107,7 +69,7 @@ class CostFunction(FileIOMixin, object):
                         "(*{}) needs explicit argument names.".format(
                             self._cost_function_handle.__name__, _par.name))
         else:
-            self._arg_names = arg_names
+            self._arg_names = list(arg_names)
         if "cost" in self._arg_names:
             raise self.__class__.EXCEPTION_TYPE(
                 "The alias 'cost' for the cost function value cannot be used as a name for one of"
@@ -121,18 +83,17 @@ class CostFunction(FileIOMixin, object):
         self._arg_count = len(self._arg_names)
         self._formatter = CostFunctionFormatter(
             name=self.name,
-            arg_formatters=[ParameterFormatter(name=_pn, value=0, error=None)
-                            for _pn in self._arg_names])
+            arg_formatters=[ParameterFormatter(_arg_name, value=0, error=None)
+                            for _arg_name in self._arg_names])
 
         self._add_constraint_cost = add_constraint_cost
         if self._add_constraint_cost:
             self._arg_names += ["parameter_values", "parameter_constraints"]
             self._arg_count += 2
 
-        self._ndf = None
         self._needs_errors = True
         self._is_chi2 = False
-        self._no_errors_warning_printed = False
+        self._saturated = False
         super(CostFunction, self).__init__()
 
     @classmethod
@@ -179,18 +140,6 @@ class CostFunction(FileIOMixin, object):
         """The :py:obj:`Formatter` objects for the function arguments"""
         return self._formatter.arg_formatters
 
-    @property
-    def ndf(self):
-        """The number of degrees of freedom of this cost function"""
-        return self._ndf
-
-    @ndf.setter
-    def ndf(self, new_ndf):
-        """The number of degrees of freedom of this cost function"""
-        assert new_ndf > 0  # ndf must be positive
-        assert new_ndf == int(new_ndf)  # ndf must be integer
-        self._ndf = int(new_ndf)
-
     @property
     def needs_errors(self):
         """Whether the cost function needs errors for a meaningful result"""
@@ -201,18 +150,49 @@ class CostFunction(FileIOMixin, object):
         """Whether the cost function is a chi2 cost function."""
         return self._is_chi2
 
-    def get_uncertainty_gaussian_approximation(self, data):
+    @property
+    def saturated(self):
+        """Whether the cost function value is calculated from a saturated likelihood."""
+        return self._saturated
+
+    def goodness_of_fit(self, *args):
+        """How well the model agrees with the data."""
+        try:
+            _index_data = self._arg_names.index(self._DATA_NAME)
+            _index_model = self._arg_names.index(self._MODEL_NAME)
+        except ValueError:
+            return None
+        _cost = self(*args)
+        args = list(args)
+        if self._add_constraint_cost:
+            args = args[:-2]
+        args[_index_model] = args[_index_data]
+        _saturated_cost = self._cost_function_handle(*args)
+        return _cost - _saturated_cost
+
+    def chi2_probability(self, cost_function_value, ndf):
+        """The chi2 probability associated with this cost function, None for non-chi2 cost
+        functions.
+
+        :param float cost_function_value: the associated cost function value.
+        :param int ndf: the associated number of degrees of freedom.
+        :returns: the associated chi2 probability.
+        :rtype: float or None
         """
-        Get the gaussian approximation of the uncertainty inherent to the cost function, returns 0
-        by default.
+        return 1.0 - chi2.cdf(cost_function_value, ndf) if self.is_chi2 else None
+
+    def get_uncertainty_gaussian_approximation(self, data):
+        """Get the gaussian approximation of the uncertainty inherent to the cost function, returns
+        0 by default.
+
         :param data: the fit data
         :return: the approximated gaussian uncertainty given the fit data
         """
         return 0
 
     def is_data_compatible(self, data):
-        """
-        Tests if model data is compatible with cost function
+        """Tests if model data is compatible with cost function
+
         :param data: the fit data
         :type data: numpy.ndarray
         :return: if the data is compatible, and if not a reason for the incompatibility
@@ -220,23 +200,19 @@ class CostFunction(FileIOMixin, object):
         """
         return True, None
 
-    def on_no_errors(self):
-        if not self._no_errors_warning_printed:
-            print('WARNING: No data errors were specified. The fit results may be wrong.')
-            self._no_errors_warning_printed = True
-
 
 class CostFunction_Chi2(CostFunction):
-    def __init__(self, errors_to_use='covariance', fallback_on_singular=True):
-        """
-        Base class for built-in least-squares cost function.
+    def __init__(
+            self, errors_to_use='covariance', fallback_on_singular=True, add_constraint_cost=True):
+        """Base class for built-in least-squares cost function.
 
-        :param errors_to_use: Which errors to use when calculating :math:`\chi^2`.
-                              Either ``'covariance'``, ``'pointwise'`` or ``None``.
+        :param errors_to_use: Which errors to use when calculating :math:`\\chi^2`.
+            Either ``'covariance'``, ``'pointwise'`` or ``None``.
         :type errors_to_use: str or None
-        :param fallback_on_singular: If :py:obj:`True` and the covariance matrix is singular (or the
-        errors are zero), calculate :math:`\chi^2` as with ``errors_to_use=None``
-        :type fallback_on_singular: bool
+        :param bool fallback_on_singular: If :py:obj:`True` and the covariance matrix is singular
+            (or the errors are zero), calculate :math:`\\chi^2` as with ``errors_to_use=None``
+        :param bool add_constraint_cost: If :py:obj:`True`, automatically add the cost for kafe2
+            constraints.
         """
 
         _cost_function_description = "chi-square"
@@ -265,13 +241,16 @@ class CostFunction_Chi2(CostFunction):
 
         super(CostFunction_Chi2, self).__init__(
             cost_function=_chi2_func,
-            arg_names=_arg_names)
+            arg_names=_arg_names,
+            add_constraint_cost=add_constraint_cost
+        )
 
         self._formatter.latex_name = "\\chi^2"
         self._formatter.name = "chi2"
         self._formatter.description = _cost_function_description
-        self._needs_errors = _chi2_func is not self.chi2_no_errors
+        self._needs_errors = errors_to_use is not None
         self._is_chi2 = True
+        self._saturated = True
 
     def _chi2(self, data, model, cov_mat_inverse=None, err=None):
         data = np.asarray(data)
@@ -303,17 +282,15 @@ class CostFunction_Chi2(CostFunction):
             else:
                 _res = _res / err
 
+        if self.needs_errors:
+            # There are other warnings that notify the user about singular cov mat, etc.
+            warnings.warn("Setting all data errors to 1 as a fallback.")
         # return sum of squared residuals
         _cost = np.sum(_res ** 2)
         if np.isnan(_cost):
             _cost = np.inf
         return _cost
 
-    def on_no_errors(self):
-        if not self._no_errors_warning_printed:
-            warnings.warn("No data errors were specified. Setting data errors to 1.")
-            self._no_errors_warning_printed = True
-
     def chi2_no_errors(self, data, model):
         r"""A least-squares cost function calculated from `y` data and model values,
         without considering uncertainties:
@@ -431,12 +408,15 @@ class CostFunction_NegLogLikelihood(CostFunction):
 
         if ratio:
             self._formatter.latex_name = r"-2\ln\mathcal{L}_{\rm R}"
-            self.formatter.name = "nllr"
+            self._formatter.name = "nllr"
         else:
-            self._formatter.latex_name = "-2\\ln\\mathcal{L}"
+            self._formatter.latex_name = r"-2\ln\mathcal{L}"
+            self._formatter.latex_name_saturated = r"-2\ln\mathcal{L}_{\rm R}"
             self._formatter.name = "nll"
+            self._formatter.name_saturated = "nllr"
         self._formatter.description = _cost_function_description
         self._needs_errors = _nll_func in [self.nll_gaussian, self.nllr_gaussian]
+        self._saturated = ratio
 
     @staticmethod
     def nll_gaussian(data, model, total_error):
@@ -445,12 +425,13 @@ class CostFunction_NegLogLikelihood(CostFunction):
         The cost function is given by:
 
         .. math::
-            C = -2 \ln \mathcal{L}({\bf d}, {\bf m}, {\bf \sigma}) = -2 \ln \prod_j \mathcal{L}_{\rm Gaussian} (x=d_j, \mu=m_j, \sigma=\sigma_j)
-                +
-                C({\bf p})
+            C = -2 \ln \mathcal{L}({\bf d}, {\bf m}, {\bf \sigma})
+              = -2 \ln \prod_j \mathcal{L}_{\rm Gaussian} (x=d_j, \mu=m_j, \sigma=\sigma_j)
+                + C({\bf p})
 
         .. math::
-            \rightarrow C = -2 \ln \prod_j \frac{1}{\sqrt{2{\sigma_j}^2\pi}} \exp{\left(-\frac{ (d_j-m_j)^2 }{ {\sigma_j}^2}\right)}
+            \rightarrow C = -2 \ln \prod_j \frac{1}{\sqrt{2{\sigma_j}^2\pi}}
+                            \exp{\left(-\frac{ (d_j-m_j)^2 }{ {\sigma_j}^2}\right)}
                             +
                             C({\bf p})
 
@@ -479,7 +460,8 @@ class CostFunction_NegLogLikelihood(CostFunction):
         The cost function is given by:
 
         .. math::
-            C = -2 \ln \mathcal{L}({\bf d}, {\bf m}) = -2 \ln \prod_j \mathcal{L}_{\rm Poisson} (k=d_j, \lambda=m_j)
+            C = -2 \ln \mathcal{L}({\bf d}, {\bf m})
+              = -2 \ln \prod_j \mathcal{L}_{\rm Poisson} (k=d_j, \lambda=m_j)
                 +
                 C({\bf p})
 
@@ -536,165 +518,6 @@ class CostFunction_NegLogLikelihood(CostFunction):
         return 0
 
 
-class CostFunctionBase_Chi2_Nuisance(CostFunction_Chi2):
-
-    def __init__(self, errors_to_use='covariance', fall_back_on_singular=True):
-        """
-        Base class for built-in least-squares cost function, which uses nuisance-parameters.
-
-          :param errors_to_use: which errors to use when calculating :math:`\chi^2`
-          :type errors_to_use: ``'covariance'``, ``'pointwise'`` or ``None``
-          :param fallback_on_singular: if ``True`` and the covariance matrix is singular (or the errors are zero),
-                                                 calculate :math:`\chi^2` as with ``errors_to_use=None``
-          :type fallback_on_singular: bool
-        """
-
-        if errors_to_use is None:
-            _chi2_nui = self.chi2_no_errors  # inherited from CostFunctionBase_Chi2
-        elif errors_to_use == 'covariance':
-            if fall_back_on_singular:
-                _chi2_nui = self.chi2_nui_cov_fallback
-            else:
-                _chi2_nui = self.chi2_nui_cov
-        elif errors_to_use == 'pointwise':
-            if fall_back_on_singular:
-                _chi2_nui = self.chi2_nui_pointwise_fallback
-            else:
-                _chi2_nui = self.chi2_nui_pointwise
-        else:
-            raise CostFunctionException(
-                "Unknown value '%s' for 'errors_to_use': must be one of ('covariance', 'pointwise', None)")
-
-        super(CostFunction_Chi2, self).__init__(cost_function=_chi2_nui)
-        #set flag for creating nuisance-parameters
-        self.set_flag("need_nuisance", True)
-
-        self._formatter.latex_name = r"\chi^{2}_{nui}"
-        self._formatter.name = "chi2_nui"
-        self._formatter.description = "chi-square (with nuisance parameters for correlated uncertainties)"
-
-    @staticmethod
-    def chi2_nui_cov(data, model, total_uncor_cov_mat_inverse, total_nuisance_cor_design_mat, nuisance_vector):
-        r"""A least-squares cost function that accounts for correlated uncertainties through nuisance parameters. The nuisance parameters are fitted.
-
-        The cost function is given by:
-
-        .. math::
-            C = \chi^2 =
-                ({\bf d} - {\bf m} - {\bf G}{\bf b})^{\top}
-                ({\bf V}^{\mathrm{uncor}})^{-1}
-                ({\bf d} - {\bf m} - {\bf G}{\bf b})
-                +
-                {\bf b}^2
-                +
-                C({\bf p})
-
-        In the above, :math:`{\bf d}` are the measurements,
-        :math:`{\bf m}` are the model predictions,
-        :math:{\bf cor_cov} is the total nuisance correlated covariance matrix,
-        :math:{\bf uncor_cov}} is the inverse of the total uncorrelated covariance matrix,
-        :math:{\bf nui} is the Nuisance-vector,
-        and :math:`C({\bf p})` is the additional cost resulting from any constrained parameters.
-
-        :param data: measurement data :math:`{\bf d}`
-        :param model: model predictions :math:`{\bf m}`
-        :param total_uncor_cov_mat_inverse: inverse :math:`({\bf V}^{\mathrm{uncor}})^{-1}` of the uncorrelated part of the total covariance matrix
-        :param total_nuisance_cor_design_mat: design matrix :math:`{\bf G}` containing correlated uncertainties
-        :param nuisance_vector: nuisance parameter vector :math:`{\bf b}`
-        :return: cost function value
-        """
-        return _generic_chi2_nuisance(
-            data=data, model=model, uncor_cov_mat_inverse=total_uncor_cov_mat_inverse,
-            nuisance_cor_design_mat=total_nuisance_cor_design_mat, nuisance_vector=nuisance_vector)
-
-    @staticmethod
-    def chi2_nui_cov_fallback(data, model, total_uncor_cov_mat_inverse, total_nuisance_cor_design_mat, nuisance_vector):
-        r"""A least-squares cost function that accounts for correlated uncertainties through nuisance parameters. The nuisance parameters are fitted.
-
-        .. TODO: describe fallback behavior
-
-        The cost function is given by:
-
-        .. math::
-            C = \chi^2 =
-                ({\bf d} - {\bf m} - {\bf G}{\bf b})^{\top}
-                ({\bf V}^{\mathrm{uncor}})^{-1}
-                ({\bf d} - {\bf m} - {\bf G}{\bf b})
-                +
-                {\bf b}^2
-                +
-                C({\bf p})
-
-        In the above, :math:`{\bf d}` are the measurements,
-        :math:`{\bf m}` are the model predictions,
-        :math:{\bf cor_cov} is the total nuisance correlated covariance matrix,
-        :math:{\bf uncor_cov}} is the inverse of the total uncorrelated covariance matrix,
-        :math:{\bf nui} is the Nuisance-vector,
-        and :math:`C({\bf p})` is the additional cost resulting from any constrained parameters.
-
-        :param data: measurement data :math:`{\bf d}`
-        :param model: model predictions :math:`{\bf m}`
-        :param total_uncor_cov_mat_inverse: inverse :math:`({\bf V}^{\mathrm{uncor}})^{-1}` of the uncorrelated part of the total covariance matrix
-        :param total_nuisance_cor_design_mat: design matrix :math:`{\bf G}` containing correlated uncertainties
-        :param nuisance_vector: nuisance parameter vector :math:`{\bf b}`
-
-        :return: cost function value
-        """
-        return _generic_chi2_nuisance(
-            data=data, model=model, uncor_cov_mat_inverse=total_uncor_cov_mat_inverse,
-            nuisance_cor_design_mat=total_nuisance_cor_design_mat, nuisance_vector=nuisance_vector)
-
-    @staticmethod
-    def chi2_nui_pointwise(data, model, total_error):
-        r"""A least-squares cost function calculated from the data and model values,
-              considering pointwise (uncorrelated) uncertainties for each data point:
-
-        .. math::
-            C = \chi^2({\bf d}, {\bf m}, {\bf \sigma}) = \sum_k \frac{d_k - m_k}{\sigma_k}
-                +
-                C({\bf p})
-
-        In the above, :math:`{\bf d}` are the measurements,
-        :math:`{\bf m}` are the model predictions,
-        :math:`{\bf \sigma}` are the pointwise total uncertainties,
-        and :math:`C({\bf p})` is the additional cost resulting from any constrained parameters.
-
-        :param data: measurement data :math:`{\bf d}`
-        :param model: model predictions :math:`{\bf m}`
-        :param total_error: total error vector :math:`{\bf \sigma}`
-
-        :return cost function value
-        """
-        return CostFunction_Chi2.chi2_pointwise_errors(
-            data=data, model=model, total_error=total_error)
-
-    @staticmethod
-    def chi2_nui_pointwise_fallback(data, model, total_error):
-        r"""A least-squares cost function calculated from the data and model values,
-              considering pointwise (uncorrelated) uncertainties for each data point:
-
-        .. TODO: describe fallback behavior
-
-        .. math::
-            C = \chi^2({\bf d}, {\bf m}, {\bf \sigma}) = \sum_k \frac{d_k - m_k}{\sigma_k}
-                +
-                C({\bf p})
-
-        In the above, :math:`{\bf d}` are the measurements,
-        :math:`{\bf m}` are the model predictions,
-        :math:`{\bf \sigma}` are the pointwise total uncertainties,
-        and :math:`C({\bf p})` is the additional cost resulting from any constrained parameters.
-
-        :param data: measurement data :math:`{\bf d}`
-        :param model: model predictions :math:`{\bf m}`
-        :param total_error: total error vector :math:`{\bf \sigma}`
-
-        :return cost function value
-        """
-        return CostFunction_Chi2.chi2_pointwise_errors_fallback(
-            data=data, model=model, total_error=total_error)
-
-
 STRING_TO_COST_FUNCTION = {
     'chi2': (CostFunction_Chi2, {}),
     'chi_2': (CostFunction_Chi2, {}),
