@@ -10,9 +10,13 @@ import six
 from .container import DataContainerBase, DataContainerException
 from ..io.file import FileIOMixin
 from ...config import kc
-from ...core import NexusFitter
+from ...core.fitters.nexus import Nexus, NexusError, Parameter
+from ...core.fitters.nexus_fitter import NexusFitter
 from ...core.constraint import GaussianMatrixParameterConstraint, GaussianSimpleParameterConstraint
+from ...core.error import CovMat
 from ...tools import print_dict_as_table
+from .._base.cost import CostFunction, STRING_TO_COST_FUNCTION
+from ..util import invert_matrix, add_in_quadrature
 
 __all__ = ["FitBase", "FitException"]
 
@@ -30,32 +34,104 @@ class FitBase(FileIOMixin, object):
 
     CONTAINER_TYPE = DataContainerBase
     MODEL_TYPE = None
+    MODEL_FUNCTION_TYPE = None
     PLOT_ADAPTER_TYPE = None
     EXCEPTION_TYPE = FitException
     RESERVED_NODE_NAMES = None
     _BASIC_ERROR_NAMES = {}
+    _STRING_TO_COST_FUNCTION = STRING_TO_COST_FUNCTION
+    _AXES = (None,)  # axes for which to for example create data nexus nodes
+    _MODEL_NAME = "model"
+    _MODEL_ERROR_NODE_NAMES = ["model_error", "model_cov_mat"]
 
-    def __init__(self):
+    def __init__(
+            self, data, model_function, cost_function, minimizer=None, minimizer_kwargs=None,
+            dynamic_error_algorithm="nonlinear"):
+        """This is a purely abstract class implementing the minimal interface required by all
+        types of fits.
+
+        :param minimizer: Name of the minimizer to use.
+        :type minimizer: str or None
+        :param minimizer_kwargs: Dictionary wit keywords for initializing the minimizer.
+        :type minimizer_kwargs: dict or None
+        :param dynamic_error_algorithm: how to handle errors that depend on the model parameters.
+        :type dynamic_error_algorithm: "nonlinear" or "iterative".
+        """
+        super(FitBase, self).__init__()
         self._data_container = None
         self._param_model = None
         self._nexus = None
         self._fitter = None
-        self._fit_param_names = None
-        self._fit_param_constraints = None
-        self._model_function = None
-        self._cost_function = None
+        self._fit_param_names = []  # names of all fit parameters
+        self._fit_param_constraints = []
         self._loaded_result_dict = None  # contains potential fit results from a file or multifit
-        super(FitBase, self).__init__()
+
+        # save minimizer, minimizer_kwargs for serialization
+        self._minimizer = minimizer
+        self._minimizer_kwargs = minimizer_kwargs
+
+        self.dynamic_error_algorithm = dynamic_error_algorithm
+        self._dynamic_error_warning_printed = False
+
+        # set/construct the model function object
+        if self.MODEL_FUNCTION_TYPE is None:
+            if model_function is not None:
+                raise ValueError(
+                    "Model function must be None if subclass does not define MODEL_FUNCTION_TYPE.")
+            self._model_function = model_function
+        else:
+            if isinstance(model_function, self.MODEL_FUNCTION_TYPE):
+                self._model_function = model_function
+            else:
+                self._model_function = self.MODEL_FUNCTION_TYPE(model_function)
+
+            # Disallow using reserved keywords as model function arguments:
+            if not self.RESERVED_NODE_NAMES.isdisjoint(
+                    set(self._model_function.signature.parameters)):
+                _invalid_args = self.RESERVED_NODE_NAMES.intersection(
+                    set(self._model_function.signature.parameters))
+                raise self.__class__.EXCEPTION_TYPE(
+                    "The following names are reserved and cannot be used as model function arguments: %r"
+                    % (_invalid_args,))
+
+        # set and validate the cost function
+        if isinstance(cost_function, CostFunction):
+            self._cost_function = cost_function
+        elif isinstance(cost_function, str):
+            try:
+                _cost_function_class, _kwargs = self._STRING_TO_COST_FUNCTION[cost_function]
+            except KeyError:
+                raise ValueError('Unknown cost function: %s' % cost_function)
+            self._cost_function = _cost_function_class(**_kwargs)
+        elif cost_function is not None:
+            self._cost_function = CostFunction(cost_function)
+            # self._validate_cost_function_raise()
+            # TODO: validate user-defined cost function? how?
+        else:
+            self._cost_function = None
+
+        # initialize the Nexus
+        self._init_nexus()
+
+        # initialize the Fitter
+        self._initialize_fitter()
+
+        # set the data after the cost_function has been set and nexus has been initialized
+        if data is not None:
+            self.data = data
 
     # -- private methods
 
-    def _add_property_to_nexus(self, prop, obj=None, name=None):
+    def _add_property_to_nexus(self, prop, obj=None, name=None, depends_on=None):
         """register a property of this object in the nexus as a function node"""
         obj = obj if obj is not None else self
-        return self._nexus.add_function(
+        _node = self._nexus.add_function(
             partial(getattr(obj.__class__, prop).fget, obj),
             func_name=name or prop
         )
+        if depends_on is not None:
+            self._nexus.add_dependency(name=_node.name, depends_on=depends_on)
+        return _node
 
     @classmethod
     def _get_base_class(cls):
@@ -65,34 +141,65 @@ class FitBase(FileIOMixin, object):
     def _get_object_type_name(cls):
         return 'fit'
 
-    def _new_data_container(self, *args, **kwargs):
-        """create a DataContainer of the right type for this fit"""
-        return self.__class__.CONTAINER_TYPE(*args, **kwargs)
-
-    def _new_parametric_model(self, *args, **kwargs):
-        """create a ParametricModel of the right type for this fit"""
-        return self.__class__.MODEL_TYPE(*args, **kwargs)
-
-    def _new_plot_adapter(self, *args, **kwargs):
-        """create a PlotAdapter of the right type for this fit"""
-        if self.__class__.PLOT_ADAPTER_TYPE is None:
-            raise NotImplementedError(
-                "No `PlotAdapter` configured for fit type '{}'!".format(
-                    self.__class__.__name__
-                ))
-        return self.__class__.PLOT_ADAPTER_TYPE(self, *args, **kwargs)
-
-    def _validate_model_function_for_fit_raise(self):
-        """make sure the supplied model function is compatible with the fit type"""
-        # disallow using reserved keywords as model function arguments
-        if not self.RESERVED_NODE_NAMES.isdisjoint(set(self._model_function.signature.parameters)):
-            _invalid_args = self.RESERVED_NODE_NAMES.intersection(set(self._model_function.signature.parameters))
-            raise self.__class__.EXCEPTION_TYPE(
-                "The following names are reserved and cannot be used as model function arguments: %r"
-                % (_invalid_args,))
-
     def _init_nexus(self):
-        pass
+        self._nexus = Nexus()
+
+        # -- fit parameters
+
+        # get names and default values of all parameters if the fit has a model function
+        if self._model_function is not None:
+            for _par_name, _par_value in six.iteritems(self._model_function.defaults_dict):
+                # create nexus node for function parameter
+                self._nexus.add(Parameter(_par_value, name=_par_name))
+
+                self._fit_param_names.append(_par_name)
+
+        self._nexus.add_function(lambda: self.parameter_values, func_name='parameter_values')
+        self._nexus.add_function(lambda: self.parameter_constraints,
+                                 func_name='parameter_constraints')
+        self._nexus.add_dependency('parameter_values', depends_on=self._fit_param_names)
+
+        # -- errors
+
+        for _axis in self._AXES:
+            _error_names = []
+            _mat_names = []
+            for _type in ("model", "data", "total"):
+                _name = '_'.join((_axis, _type)) if _axis is not None else _type
+                _error_name = '_'.join((_axis, _type, 'error')) if _axis is not None \
+                    else '_'.join((_type, 'error'))
+                _mat_name = '_'.join((_axis, _type, "cov_mat")) if _axis is not None \
+                    else '_'.join((_type, "cov_mat"))
+                if _type == "total":
+                    self._add_property_to_nexus(_error_name, depends_on=_error_names)
+                    self._add_property_to_nexus(_mat_name, depends_on=_mat_names)
+                else:
+                    self._add_property_to_nexus(_name)
+                    self._add_property_to_nexus(_error_name)
+                    _error_names.append(_error_name)
+                    self._add_property_to_nexus(_mat_name)
+                    _mat_names.append(_mat_name)
+                if _type == "model":
+                    self._nexus.add_dependency(_error_name, depends_on="parameter_values")
+                    self._nexus.add_dependency(_mat_name, depends_on="parameter_values")
+                self._add_property_to_nexus(_mat_name + "_inverse", depends_on=_mat_name)
+
+        if self._model_function is not None:
+            # add the original function name as an alias to 'model'
+            try:
+                self._nexus.add_alias(self._model_function.name, alias_for=self._MODEL_NAME)
+            except NexusError:
+                pass  # allow 'model' as function name for model
+
+        if self._cost_function is not None:
+            # the cost function (the function to be minimized)
+            _cost_node = self._nexus.add_function(
+                self._cost_function,
+                par_names=self._cost_function.arg_names,
+                func_name=self._cost_function.name,
+            )
+
+            _cost_alias = self._nexus.add_alias('cost', alias_for=self._cost_function.name)
 
     def _initialize_fitter(self):
         self._fitter = NexusFitter(nexus=self._nexus,
@@ -101,27 +208,6 @@ class FitBase(FileIOMixin, object):
                                    minimizer=self._minimizer,
                                    minimizer_kwargs=self._minimizer_kwargs)
 
-    @staticmethod
-    def _get_default_values(model_function=None, x_name=None):
-        """Gets the default values for all independent parameter names. If no default is present in the function
-            definition, the default from the kafe2 config is loaded.
-        :param model_function: model function handle
-        :param str x_name: name of the independent parameter
-        :return: Ordered dict with default values for fit parameters.
-        :rtype: OrderedDict[str, float]
-        """
-        _nexus_new_dict = OrderedDict()
-        for _par in model_function.signature.parameters.values():
-            # skip independent variable parameter
-            if _par.name == x_name:
-                continue
-            if _par.default == _par.empty:
-                _nexus_new_dict[_par.name] = kc('core', 'default_initial_parameter_value')
-            else:
-                _nexus_new_dict[_par.name] = _par.default
-
-        return _nexus_new_dict
-
     @abc.abstractmethod
     def _set_new_data(self, new_data):
         """Private method called by the :py:attr:`~data` setter. Must be overwritten for each derived class."""
@@ -131,12 +217,12 @@ class FitBase(FileIOMixin, object):
     def _set_new_parametric_model(self):
         pass
 
-    # Overwritten by MultiFit
+    # Overridden by MultiFit
     def _get_model_function_argument_formatters(self):
         """All arguments of the model function including independent variables."""
         return self._model_function.formatter.arg_formatters
 
-    # Overwritten by MultiFit
+    # Overridden by MultiFit
     def _get_model_function_parameter_formatters(self):
         """Only the parameters which are fitted. Excludes independent variables."""
         return self.model_function.formatter.par_formatters
@@ -204,10 +290,12 @@ class FitBase(FileIOMixin, object):
 
         _cor_mat_content = self.parameter_cor_mat
         if _cor_mat_content is not None:
+            par_display_names = [_pf.name for _pf in
+                                 self._get_model_function_parameter_formatters()]
             _cor_mat_as_dict = OrderedDict()
-            _cor_mat_as_dict['_invisible_first_column'] = self.poi_names
-            for _poi_name, _row in zip(self.poi_names, self.parameter_cor_mat.T):
-                _cor_mat_as_dict[_poi_name] = np.atleast_1d(np.squeeze(np.asarray(_row)))
+            _cor_mat_as_dict['_invisible_first_column'] = par_display_names
+            for _par_name, _row in zip(par_display_names, self.parameter_cor_mat.T):
+                _cor_mat_as_dict[_par_name] = np.atleast_1d(np.squeeze(np.asarray(_row)))
 
             print_dict_as_table(_cor_mat_as_dict, output_stream=output_stream, indent_level=2)
         else:
@@ -218,16 +306,35 @@ class FitBase(FileIOMixin, object):
         output_stream.write(indent * (indentation_level + 1) + "=============\n\n")
 
         _pf = self._cost_function.formatter
-        output_stream.write(indent * (indentation_level + 2) + "cost function: {}\n\n".format(_pf.description))
-        output_stream.write(indent * (indentation_level + 2) + "cost / ndf = ")
         output_stream.write(
-            _pf.get_formatted(value=self.cost_function_value,
-                              n_degrees_of_freedom=self.ndf,
-                              with_name=False,
-                              with_value_per_ndf=True,
-                              format_as_latex=False)
-        )
-        output_stream.write('\n')
+            indent * (indentation_level + 2) + "cost function: {}\n\n".format(_pf.description))
+        if self._cost_function.saturated:
+            _gof_string = "cost / ndf = "
+            _gof_value = self.cost_function_value
+        else:
+            output_stream.write(indent * (indentation_level + 2) + "cost = ")
+            output_stream.write(_pf.get_formatted(
+                value=self.cost_function_value,
+                with_name=False,
+                format_as_latex=False,
+            ))
+            output_stream.write('\n\n')
+            _gof_string = "GoF / ndf = "
+            _gof_value = self.goodness_of_fit
+        if _gof_value is not None:
+            output_stream.write(indent * (indentation_level + 2) + _gof_string)
+            output_stream.write(_pf.get_formatted(
+                value=_gof_value,
+                n_degrees_of_freedom=self.ndf,
+                with_name=False,
+                with_value_per_ndf=True,
+                format_as_latex=False
+            ))
+            output_stream.write('\n\n')
+        _chi2_prob = self.chi2_probability
+        if _chi2_prob is not None:
+            output_stream.write("%schi2 probability = %#.3g\n\n" % (
+                indent * (indentation_level + 2), _chi2_prob))
 
     def _update_parameter_formatters(self, update_asymmetric_errors=False):
         """Update all parameter formatters with the current values and uncertainties.
@@ -239,6 +346,7 @@ class FitBase(FileIOMixin, object):
             _fpf.value = _pv
             _fpf.error = _pe
         if update_asymmetric_errors:
+            self._check_dynamic_error_compatibility()
             for _fpf, _ape in zip(self._get_model_function_parameter_formatters(), self.asymmetric_parameter_errors):
                 _fpf.asymmetric_error = _ape
 
@@ -248,16 +356,52 @@ class FitBase(FileIOMixin, object):
         for _error_name in self._BASIC_ERROR_NAMES:
             self._nexus.get(_error_name).mark_for_update()
 
+    def _set_data_as_model_ref(self):
+        for _err in self._param_model.get_matching_errors({"relative": True}).values():
+            _old_ref = _err.reference
+            _err.reference = self._data_container.data
+
+    def _iterative_fits_needed(self):
+        return bool(self._param_model.get_matching_errors({"relative": True})) \
+               and self._dynamic_error_algorithm == "iterative"
+
+    def _second_fit_needed(self):
+        return bool(self._param_model.get_matching_errors({"relative": True})) \
+               and self._dynamic_error_algorithm == "nonlinear"
+
+    def _get_node_names_to_freeze(self, first_fit):
+        if first_fit or not self._param_model.get_matching_errors({"relative": True}) \
+                or self._dynamic_error_algorithm == "iterative":
+            return self._MODEL_ERROR_NODE_NAMES
+        else:
+            return []
+
+    def _pre_fit_iteration(self, first_fit=False):
+        for _model_err_name in self._get_node_names_to_freeze(first_fit):
+            _node = self._nexus.get(_model_err_name)
+            _node.update()
+            _node.freeze()
+
+    def _post_fit_iteration(self, first_fit=False):
+        for _model_err_name in self._get_node_names_to_freeze(first_fit):
+            _node = self._nexus.get(_model_err_name)
+            _node.unfreeze()
+            _node.update()
+            _node.notify_parents()
+
+    def _check_dynamic_error_compatibility(self):
+        if not self._dynamic_error_warning_printed and self._iterative_fits_needed():
+            warnings.warn(
+                "Asymmetric parameter errors, parameter profiles, and contours cannot be "
+                "calculated with iterative dynamic errors. Used nonlinear errors instead.")
+            self._dynamic_error_warning_printed = True
+
     # -- public properties
 
     @property
-    @abc.abstractmethod
     def data(self):
-        """The data used in this fit.
-
-        :rtype: numpy.ndarray
-        """
-        pass
+        """array of measurement values"""
+        return self._data_container.data
 
     @data.setter
     def data(self, new_data):
@@ -267,9 +411,27 @@ class FitBase(FileIOMixin, object):
         if not _data_and_cost_compatible:
             raise self.EXCEPTION_TYPE('Fit data and cost function are not compatible: %s' % _reason)
         self._set_new_parametric_model()
-        # TODO: check where to update this (set/release/etc.)
-        # FIXME: nicer way than len()?
-        self._cost_function.ndf = self._data_container.size - len(self._param_model.parameters)
+        self._param_model._on_error_change_callbacks = [self._on_error_change]
+
+    @property
+    def data_error(self):
+        """array of pointwise data uncertainties"""
+        return self._data_container.err
+
+    @property
+    def data_cov_mat(self):
+        """the data covariance matrix"""
+        return self._data_container.cov_mat
+
+    @property
+    def data_cov_mat_inverse(self):
+        """inverse of the data covariance matrix (or ``None`` if singular)"""
+        return self._data_container.cov_mat_inverse
+
+    @property
+    def data_cor_mat(self):
+        """the data correlation matrix"""
+        return self._data_container.cor_mat
 
     @property
     def data_container(self):
@@ -284,32 +446,49 @@ class FitBase(FileIOMixin, object):
     def model(self):
         pass
 
-    # @abc.abstractproperty
-    # def data_error(self): pass
+    @property
+    def model_error(self):
+        """array of pointwise model uncertainties"""
+        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
+        return self._param_model.err
 
-    # @abc.abstractproperty
-    # def data_cov_mat(self): pass
-    #
-    # @abc.abstractproperty
-    # def data_cov_mat_inverse(self): pass
-    #
-    # @abc.abstractproperty
-    # def model_error(self): pass
-    #
-    # @abc.abstractproperty
-    # def model_cov_mat(self): pass
-    #
-    # @abc.abstractproperty
-    # def model_cov_mat_inverse(self): pass
-    #
-    # @abc.abstractproperty
-    # def total_error(self): pass
-    #
-    # @abc.abstractproperty
-    # def total_cov_mat(self): pass
-    #
-    # @abc.abstractproperty
-    # def total_cov_mat_inverse(self): pass
+    @property
+    def model_cov_mat(self):
+        """the model covariance matrix"""
+        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
+        return self._param_model.cov_mat
+
+    @property
+    def model_cov_mat_inverse(self):
+        """inverse of the model covariance matrix (or ``None`` if singular)"""
+        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
+        return self._param_model.cov_mat_inverse
+
+    @property
+    def model_cor_mat(self):
+        """the model correlation matrix"""
+        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
+        return self._param_model.cor_mat
+
+    @property
+    def total_error(self):
+        """array of pointwise total uncertainties"""
+        return add_in_quadrature(self.data_error, self.model_error)
+
+    @property
+    def total_cov_mat(self):
+        """the total covariance matrix"""
+        return self.data_cov_mat + self.model_cov_mat
+
+    @property
+    def total_cov_mat_inverse(self):
+        """inverse of the total covariance matrix (or ``None`` if singular)"""
+        return invert_matrix(self.total_cov_mat)
+
+    @property
+    def total_cor_mat(self):
+        """the total correlation matrix"""
+        return CovMat(self.total_cov_mat).cor_mat
 
     @property
     def model_function(self):
@@ -387,6 +566,7 @@ class FitBase(FileIOMixin, object):
         """
         if self._loaded_result_dict is not None and self._loaded_result_dict['asymmetric_parameter_errors'] is not None:
             return self._loaded_result_dict['asymmetric_parameter_errors']
+        self._check_dynamic_error_compatibility()
         return self._fitter.asymmetric_fit_parameter_errors
 
     @property
@@ -454,23 +634,6 @@ class FitBase(FileIOMixin, object):
         """
         return 1
 
-    @property
-    def poi_values(self):
-        """The values of the parameters of interest, equal to :py:attr:`~parameter_values` minus nuisance parameters.
-
-        :rtype: numpy.ndarray[float]
-        """
-        return self.parameter_values
-
-    @property
-    def poi_names(self):
-        """The names of the parameters of interest, equal to :py:attr:`~parameter_names` minus nuisance parameter
-        names
-
-        :rtype: tuple[str]
-        """
-        return self.parameter_names
-
     @property
     def did_fit(self):
         """Whether a fit was performed for the given data and model.
@@ -487,7 +650,33 @@ class FitBase(FileIOMixin, object):
 
         :rtype: int
         """
-        return self._cost_function.ndf + len(self._fitter.fixed_parameters)
+        return self._param_model.ndf + len(self._fitter.fixed_parameters)
+
+    @property
+    def goodness_of_fit(self):
+        return self._cost_function.goodness_of_fit(
+            *[self._nexus.get(_node_name).value for _node_name in self._cost_function.arg_names])
+
+    @property
+    def dynamic_error_algorithm(self):
+        """The algorithm to use for handling errors that depend on the model parameters.
+        :rtype: str
+        """
+        return self._dynamic_error_algorithm
+
+    @dynamic_error_algorithm.setter
+    def dynamic_error_algorithm(self, new_dea):
+        _valid_deas = ["nonlinear", "iterative"]
+        if new_dea not in _valid_deas:
+            raise ValueError(
+                "Unknown dynamic error algorithm: %s. Valid algorithms: %s" % (
+                    new_dea, _valid_deas))
+        self._dynamic_error_algorithm = new_dea
+
+    @property
+    def chi2_probability(self):
+        """The chi2 probability for the current model values."""
+        return self._cost_function.chi2_probability(self.cost_function_value, self.ndf)
 
     # -- public methods
 
@@ -587,7 +776,7 @@ class FitBase(FileIOMixin, object):
         _par_indices = []
         for _name in names:
             try:
-                _par_indices.append(self.poi_names.index(_name))
+                _par_indices.append(self.parameter_names.index(_name))
             except ValueError:
                 raise self.EXCEPTION_TYPE('Unknown parameter name: %s' % _name)
         self._fit_param_constraints.append(GaussianMatrixParameterConstraint(
@@ -604,7 +793,7 @@ class FitBase(FileIOMixin, object):
         :param bool relative: Whether the given uncertainty is relative to the given value.
         """
         try:
-            _index = self.poi_names.index(name)
+            _index = self.parameter_names.index(name)
         except ValueError:
             raise self.EXCEPTION_TYPE('Unknown parameter name: %s' % name)
         self._fit_param_constraints.append(GaussianSimpleParameterConstraint(
@@ -686,8 +875,6 @@ class FitBase(FileIOMixin, object):
         elif reference == 'model':
             # delegate to model container
             _reference_object = self._param_model
-            if relative:
-                raise NotImplementedError("Errors relative to model not implemented!")
         else:
             raise FitException("Cannot add error: unknown reference "
                                "specification '{}', expected one of: 'data', 'model'...".format(reference))
@@ -768,13 +955,45 @@ class FitBase(FileIOMixin, object):
         :return: A dictionary containing the fit results.
         :rtype: dict
         """
-        if self._cost_function.needs_errors and not self._data_container.has_errors:
-            self._cost_function.on_no_errors()
+        if self._cost_function.needs_errors and not self.has_errors:
+            warnings.warn("Cost function expects errors but no errors were specified.")
+
+        # Give relative model errors data as reference for initial fit:
+        self._set_data_as_model_ref()
+
+        # Initial fit:
+        self._pre_fit_iteration(first_fit=True)
         self._fitter.do_fit()  # TODO specify other node to minimize
+        self._post_fit_iteration(first_fit=True)
+
+        if self._iterative_fits_needed():
+            _convergence_limit = float(kc("fit", "iterative_do_fit", "convergence_limit"))
+            _previous_cost = self.cost_function_value
+            for i in range(kc("fit", "iterative_do_fit", "max_iterations")):
+                self._pre_fit_iteration()
+                self._fitter.reset_minimizer()  # flush iminuit cache
+                self._fitter.do_fit()
+                self._post_fit_iteration()
+                if abs(self.cost_function_value - _previous_cost) < _convergence_limit:
+                    break
+                _previous_cost = self.cost_function_value
+        elif self._second_fit_needed():
+            self._pre_fit_iteration()
+            self._fitter.reset_minimizer()  # flush iminuit cache
+            self._fitter.do_fit()
+            self._post_fit_iteration()
+
         self._loaded_result_dict = None
         self._update_parameter_formatters()
         return self.get_result_dict(asymmetric_parameter_errors=asymmetric_parameter_errors)
 
+    def assign_model_function_name(self, name):
+        """Assign a string to be the model function name.
+
+        :param str name: The new name.
+        """
+        self._model_function.formatter.name = name
+
     def assign_model_function_expression(self, expression_format_string):
         """Assign a plain-text-formatted expression string to the model function.
 
@@ -782,6 +1001,19 @@ class FitBase(FileIOMixin, object):
         """
         self._model_function.formatter.expression_format_string = expression_format_string
 
+    def assign_parameter_names(self, **par_names_dict):
+        """Assign display strings to all model function arguments.
+
+        :param par_names_dict: Dictionary mapping the parameter names to their display names.
+        """
+        for _af in self._get_model_function_argument_formatters():
+            _an = par_names_dict.pop(_af.arg_name, None)
+            if _an is not None:
+                _af.name = _an
+        if par_names_dict:
+            warnings.warn("Could not assign all names to a parameter."
+                          "Leftover: {}".format(par_names_dict))
+
     def assign_model_function_latex_name(self, latex_name):
         """Assign a LaTeX-formatted string to be the model function name.
 
@@ -792,11 +1024,12 @@ class FitBase(FileIOMixin, object):
     def assign_model_function_latex_expression(self, latex_expression_format_string):
         """Assign a LaTeX-formatted expression string to the model function.
 
-        :param str latex_expression_format_string: The LaTeX string. Elements like ``'{par_name}'`` will be replaced
-            automatically with the corresponding LaTeX names for the given parameter. These can be set with
-            :py:meth:`~assign_parameter_latex_names`.
+        :param str latex_expression_format_string: The LaTeX string. Elements like ``'{par_name}'``
+            will be replaced automatically with the corresponding LaTeX names for the given
+            parameter. These can be set with :py:meth:`~assign_parameter_latex_names`.
         """
-        self._model_function.formatter.latex_expression_format_string = latex_expression_format_string
+        self._model_function.formatter.latex_expression_format_string = \
+            latex_expression_format_string
 
     def assign_parameter_latex_names(self, **par_latex_names_dict):
         """Assign LaTeX-formatted strings to all model function arguments.
@@ -804,11 +1037,12 @@ class FitBase(FileIOMixin, object):
         :param par_latex_names_dict: Dictionary mapping the parameter names to their latex names.
         """
         for _af in self._get_model_function_argument_formatters():
-            _aln = par_latex_names_dict.pop(_af.name, None)
+            _aln = par_latex_names_dict.pop(_af.arg_name, None)
             if _aln is not None:
                 _af.latex_name = _aln
         if par_latex_names_dict:
-            warnings.warn("Could not assign all latex names to a parameter. Leftover: {}".format(par_latex_names_dict))
+            warnings.warn("Could not assign all latex names to a parameter."
+                          "Leftover: {}".format(par_latex_names_dict))
 
     def get_result_dict(self, asymmetric_parameter_errors=False):
         """Return a dictionary of the fit results.
@@ -824,7 +1058,9 @@ class FitBase(FileIOMixin, object):
         _ndf = self.ndf
         _result_dict['cost'] = _cost
         _result_dict['ndf'] = _ndf
+        _result_dict['goodness_of_fit'] = self.goodness_of_fit
         _result_dict['cost/ndf'] = _cost / _ndf
+        _result_dict['chi2_probability'] = self.chi2_probability
         _result_dict['parameter_values'] = self.parameter_name_value_dict
         if _result_dict['did_fit']:
             _result_dict['parameter_cov_mat'] = self.parameter_cov_mat
@@ -841,6 +1077,7 @@ class FitBase(FileIOMixin, object):
         if self._loaded_result_dict is not None and self._loaded_result_dict['asymmetric_parameter_errors'] is not None:
             _asymm_errs = self._loaded_result_dict['asymmetric_parameter_errors']
         elif asymmetric_parameter_errors:
+            self._check_dynamic_error_compatibility()
             _asymm_errs = self.asymmetric_parameter_errors
         else:
             _asymm_errs = self._fitter.asymmetric_fit_parameter_errors_if_calculated
