@@ -5,6 +5,7 @@ import textwrap
 import warnings
 import itertools
 import matplotlib as mpl
+import sys
 
 from ..multi.fit import MultiFit
 from ...config import kc, ConfigError, kafe2_rc
@@ -18,6 +19,9 @@ from matplotlib import rc_context
 __all__ = ["PlotAdapterBase", "Plot", "PlotAdapterException", "PlotFigureException",
            "kc_plot_style"]
 
+_python_version = sys.version_info[0]
+_float_template = ".3g" if _python_version == 2 else "#.3g"
+
 
 def kc_plot_style(data_type, subplot_key, property_key):
     try:
@@ -558,10 +562,22 @@ class Plot(object):
     """
     # TODO update documentation
 
-    FIT_INFO_STRING_FORMAT = textwrap.dedent("""\
+    FIT_INFO_STRING_FORMAT_CHI2 = textwrap.dedent("""\
         {model_function}
         {parameters}
             $\\hookrightarrow${fit_quality}
+            $\\hookrightarrow \\chi^2 \\, \\mathrm{{probability =}}${chi2_probability:%s}
+    """ % _float_template)
+    FIT_INFO_STRING_FORMAT_SATURATED = textwrap.dedent("""\
+        {model_function}
+        {parameters}
+            $\\hookrightarrow${fit_quality}
+    """)
+    FIT_INFO_STRING_FORMAT_NOT_SATURATED = textwrap.dedent("""\
+        {model_function}
+        {parameters}
+            $\\hookrightarrow${cost}
+            $\\hookrightarrow${fit_quality}
     """)
 
     def __init__(self, fit_objects, separate_figures=False):
@@ -638,16 +654,23 @@ class Plot(object):
             self._plot_adapters = []
             for _fit in self._fits:
                 self._plot_adapters.append(
-                    _fit._new_plot_adapter()
+                    _fit.PLOT_ADAPTER_TYPE(_fit)
                 )
 
         return [self._plot_adapters[_idx] for _idx in plot_indices]
 
     def _plot_and_get_results(self, plot_indices=None):
         plot_indices = plot_indices or range(len(self._fits))
-        for _fit in self._fits:
-            if not _fit.did_fit:
-                warnings.warn("No fit has been performed for {}. Did you forget to run fit.do_fit()?".format(_fit))
+        if self._multifit is None:
+            for _fit in self._fits:
+                if not _fit.did_fit:
+                    warnings.warn(
+                        "No fit has been performed for {}. Did you forget to run fit.do_fit()?"
+                            .format(_fit))
+        elif not self._multifit.did_fit:
+            warnings.warn(
+                "No fit has been performed for {}. Did you forget to run fit.do_fit()?"
+                    .format(self._multifit))
 
         _plot_adapters = self._get_plot_adapters(plot_indices)
 
@@ -718,7 +741,11 @@ class Plot(object):
 
         _ndf = plot_adapter._fit.ndf
         _cost_function_value = plot_adapter._fit.cost_function_value
-        _info_text = self.FIT_INFO_STRING_FORMAT.format(
+        if plot_adapter._fit._cost_function.saturated:
+            _gof_value = _cost_function_value
+        else:
+            _gof_value = plot_adapter._fit.goodness_of_fit
+        _info_format_dict = dict(
             model_function=plot_adapter.get_formatted_model_function(
                 with_par_values=False,
                 n_significant_digits=2,
@@ -734,24 +761,76 @@ class Plot(object):
                     format_as_latex=format_as_latex
                 )
                 for _pf in plot_adapter.model_function_parameter_formatters
-            ]),
-            fit_quality=_cost_func._formatter.get_formatted(
-                value=_cost_function_value,
+            ])
+        )
+        # _gof_value is None for UnbinnedFit and some user-defined cost functions.
+        if _gof_value is not None:
+            _info_format_dict["fit_quality"] = _cost_func.formatter.get_formatted(
+                value=_gof_value,
+                with_name=True,
+                saturated=True,
                 n_degrees_of_freedom=_ndf,
                 with_value_per_ndf=True,
                 format_as_latex=format_as_latex
-            ),
-        )
+            )
+            if plot_adapter._fit._cost_function.is_chi2:
+                _info_format_string = self.FIT_INFO_STRING_FORMAT_CHI2
+                _info_format_dict["chi2_probability"] = plot_adapter._fit.chi2_probability
+            elif plot_adapter._fit._cost_function.saturated:
+                _info_format_string = self.FIT_INFO_STRING_FORMAT_SATURATED
+            else:
+                _info_format_string = self.FIT_INFO_STRING_FORMAT_NOT_SATURATED
+                _info_format_dict["cost"] = _cost_func.formatter.get_formatted(
+                    value=_cost_function_value,
+                    with_name=True,
+                    format_as_latex=format_as_latex
+                )
+        else:
+            _info_format_string = self.FIT_INFO_STRING_FORMAT_SATURATED
+            _info_format_dict["fit_quality"] = _cost_func.formatter.get_formatted(
+                value=_cost_function_value,
+                with_name=True,
+                format_as_latex=format_as_latex
+            )
+
+        _info_text = _info_format_string.format(**_info_format_dict)
+
         if self._multifit is not None:
             _multi_ndf = self._multifit.ndf
+            _multi_cost_function = self._multifit._cost_function
             _multi_cost_function_value = self._multifit.cost_function_value
-            _info_text += "    $\\hookrightarrow$ global {}".format(
-                _cost_func._formatter.get_formatted(
+
+            _multi_info_dict = dict()
+            if _multi_cost_function.is_chi2:
+                _multi_gof = _multi_cost_function_value
+                _template = "    $\\hookrightarrow$ global {fit_quality}\n"
+                _multi_info_dict["chi2_probability"] = self._multifit.chi2_probability
+                _template += "    $\\hookrightarrow$ global $\\chi^2 \\, \\mathrm{{probability}} " \
+                             "= {chi2_probability:%s}$" % _float_template
+            elif _multi_cost_function.saturated:
+                _multi_gof = _multi_cost_function_value
+                _template = "    $\\hookrightarrow$ global cost / ndf = {fit_quality}\n"
+            else:
+                _template = "    $\\hookrightarrow$ global cost = {cost}\n"
+                _multi_info_dict["cost"] = _multi_cost_function.formatter.get_formatted(
                     value=_multi_cost_function_value,
+                    with_name=False,
+                    format_as_latex=format_as_latex
+                )
+                _multi_gof = self._multifit.goodness_of_fit
+                if _multi_gof is not None:
+                    _template += "    $\\hookrightarrow$ global GoF / ndf = {fit_quality}\n"
+
+            if _multi_gof is not None:
+                _multi_info_dict["fit_quality"] = _multi_cost_function.formatter.get_formatted(
+                    value=_multi_gof,
+                    with_name=_multi_cost_function.is_chi2,
                     n_degrees_of_freedom=_multi_ndf,
                     with_value_per_ndf=True,
                     format_as_latex=format_as_latex
-                ))
+                )
+
+            _info_text += _template.format(**_multi_info_dict)
         return _info_text
 
     def _render_legend(self, plot_results, axes_keys, fit_info=True, asymmetric_parameter_errors=False, **kwargs):
