@@ -12,7 +12,8 @@ class SharedCostFunction(CostFunction_Chi2):
         self._MODEL_NAME = "y_model"
         self._COV_MAT_INVERSE_NAME = "total_cov_mat_inverse"
         super(SharedCostFunction, self).__init__(
-            errors_to_use="covariance", fallback_on_singular=fallback_on_singular)
+            errors_to_use="covariance", fallback_on_singular=fallback_on_singular,
+            add_constraint_cost=False)
 
 
 class MultiCostFunction(CostFunction):
@@ -23,11 +24,15 @@ class MultiCostFunction(CostFunction):
             arg_names=cost_function_names,
             add_constraint_cost=True
         )
-        self._needs_errors = False
-        for _singular_cost_function in singular_cost_functions:
-            if _singular_cost_function.needs_errors:
-                self._needs_errors = True
-                break
+        self._needs_errors = np.any([_scf.needs_errors for _scf in singular_cost_functions])
+        self._saturated = np.all([_scf.saturated for _scf in singular_cost_functions])
+        self._is_chi2 = np.all([_scf.is_chi2 for _scf in singular_cost_functions])
+        if self.is_chi2:
+            self.formatter.name = "chi2"
+            self.formatter.latex_name = "\\chi^2"
+        else:
+            self.formatter.name = "cost"
+            self.formatter.latex_name = None
 
     @staticmethod
     def cost_sum(*single_costs):
