@@ -92,64 +92,62 @@ class TestIndexedFitBasicInterface(AbstractTestFit, unittest.TestCase):
 
         # helper dict with all reference property values
         self._ref_prop_dict = dict(
-            parameter_values=self._ref_initial_pars,
-            poi_values=self._ref_initial_pars,
-
-            data=self._ref_data,
-            model=self._ref_initial_model,
-            poi_names=('a', 'b', 'c'),
             did_fit=False,
             model_count=1,
+
+            parameter_values=self._ref_initial_pars,
+            parameter_names=('a', 'b', 'c'),
             cost_function_value=self._ref_initial_cost,
 
-            data_cov_mat=self._ref_matrix_eye,
-            #data_uncor_cov_mat=self._ref_matrix_eye,
-            data_cov_mat_inverse=self._ref_matrix_eye,
-            #data_uncor_cov_mat_inverse=self._ref_matrix_eye,
-            data_cor_mat=self._ref_matrix_eye,
+            data=self._ref_data,
             data_error=self._ref_error,
+            data_cov_mat=self._ref_matrix_eye,
+            data_cov_mat_inverse=self._ref_matrix_eye,
+            data_cor_mat=self._ref_matrix_eye,
 
-            model_cov_mat=self._ref_matrix_eye * 0,
-            #model_uncor_cov_mat=self._ref_matrix_eye * 0,  # TODO: fix
-            model_cov_mat_inverse=None,
-            #model_uncor_cov_mat_inverse=None,  # TODO: fix
-            #model_cor_mat=self._ref_matrix_eye,
+            model=self._ref_initial_model,
             model_error=self._ref_error * 0,
+            model_cov_mat=self._ref_matrix_eye * 0,
+            model_cov_mat_inverse=None,
+            model_cor_mat=self._ref_matrix_eye * np.nan,
 
-            total_cov_mat=self._ref_matrix_eye,
-            #total_uncor_cov_mat=self._ref_matrix_eye,
-            total_cov_mat_inverse=self._ref_matrix_eye,
-            #total_uncor_cov_mat_inverse=self._ref_matrix_eye,
-            #total_cor_mat=self._ref_matrix_eye,  # TODO: add
             total_error=self._ref_error,
+            total_cov_mat=self._ref_matrix_eye,
+            total_cov_mat_inverse=self._ref_matrix_eye,
+            total_cor_mat=self._ref_matrix_eye,
         )
 
-    def _get_fit(self, model_function=None, cost_function=None, error=None):
+    def _get_fit(
+            self, model_function=None, cost_function=None, errors=None,
+            dynamic_error_algorithm=None):
         '''convenience'''
         model_function = model_function or simple_indexed_model
         # TODO: fix default
         cost_function = cost_function or IndexedCostFunction_Chi2(
             errors_to_use='covariance')
-        error = error or 1.0
+        errors = errors or [dict(err_val=1.0)]
+        dynamic_error_algorithm = dynamic_error_algorithm or "nonlinear"
 
         _fit = IndexedFit(
             data=self._ref_data,
             model_function=model_function,
             cost_function=cost_function,
-            minimizer=self.MINIMIZER
+            minimizer=self.MINIMIZER,
+            dynamic_error_algorithm=dynamic_error_algorithm
         )
-        _fit.add_error(err_val=error)
+        for _err in errors:
+            _fit.add_error(**_err)
 
         return _fit
 
     def _get_test_fits(self):
         return {
-            'default': \
-                self._get_fit(),
-            'explicit': \
-                self._get_fit(cost_function=simple_chi2),
-            'explicit_model': \
-                self._get_fit(cost_function=simple_chi2_explicit_model_name),
+            'default': self._get_fit(),
+            'explicit': self._get_fit(cost_function=simple_chi2),
+            'explicit_model': self._get_fit(cost_function=simple_chi2_explicit_model_name),
+            'relative_errors_data': self._get_fit(
+                errors=[dict(err_val=1.0 / self._ref_data, relative=True, reference='data')]
+            ),
         }
 
     def test_initial_state(self):
@@ -161,7 +159,6 @@ class TestIndexedFitBasicInterface(AbstractTestFit, unittest.TestCase):
         self.run_test_for_all_fits(
             dict(self._ref_prop_dict,
                 parameter_values=self._nominal_fit_result_pars,
-                poi_values=self._nominal_fit_result_pars,
                 model=self._nominal_fit_result_model,
                 did_fit=True,
                 cost_function_value=self._nominal_fit_result_cost,
@@ -174,10 +171,9 @@ class TestIndexedFitBasicInterface(AbstractTestFit, unittest.TestCase):
         self.run_test_for_all_fits(
             dict(self._ref_prop_dict,
                 parameter_values=self._nominal_fit_result_pars,
-                poi_values=self._nominal_fit_result_pars,
                 data=self._ref_data,
                 model=self._nominal_fit_result_model,
-                poi_names=('a', 'b', 'c'),
+                parameter_names=('a', 'b', 'c'),
                 did_fit=False,
                 model_count=1,
                 cost_function_value=self._nominal_fit_result_cost
@@ -206,7 +202,6 @@ class TestIndexedFitBasicInterface(AbstractTestFit, unittest.TestCase):
             _fit,
             dict(
                 parameter_values=np.array([1, 1, 1]),
-                poi_values=np.array([1, 1, 1]),
             )
         )
 
@@ -220,7 +215,6 @@ class TestIndexedFitBasicInterface(AbstractTestFit, unittest.TestCase):
             _fit,
             dict(
                 parameter_values=np.array([1, 1, 3.3]),
-                poi_values=np.array([1, 1, 3.3]),
             )
         )
 
@@ -245,7 +239,6 @@ class TestIndexedFitBasicInterface(AbstractTestFit, unittest.TestCase):
             dict(
                 data=self._ref_data * 2,
                 parameter_values=_new_estimates,
-                poi_values=_new_estimates,
             ),
             rtol=1e-2
         )
@@ -337,6 +330,32 @@ class TestIndexedFitBasicInterface(AbstractTestFit, unittest.TestCase):
         _fit.report(output_stream=_buffer)
         self.assertNotEqual(_buffer.getvalue(), "")
 
+    def test_relative_model_error_nonlinear(self):
+        _fit_data_err = self._get_fit(errors=[
+            dict(err_val=1.0, relative=False, reference="data"),
+            dict(err_val=0.1, relative=True, reference="data")
+        ])
+        _fit_model_err = self._get_fit(errors=[
+            dict(err_val=1.0, relative=False, reference="data"),
+            dict(err_val=0.1, relative=True, reference="model")
+        ], dynamic_error_algorithm="nonlinear")
+        _fit_data_err.do_fit()
+        _fit_model_err.do_fit()
+        self._assert_fit_results_equal(_fit_data_err, _fit_model_err, rtol=3e-2)
+
+    def test_relative_model_error_iterative(self):
+        _fit_data_err = self._get_fit(errors=[
+            dict(err_val=1.0, relative=False, reference="data"),
+            dict(err_val=0.1, relative=True, reference="data")
+        ])
+        _fit_model_err = self._get_fit(errors=[
+            dict(err_val=1.0, relative=False, reference="data"),
+            dict(err_val=0.1, relative=True, reference="model")
+        ], dynamic_error_algorithm="iterative")
+        _fit_data_err.do_fit()
+        _fit_model_err.do_fit()
+        self._assert_fit_results_equal(_fit_data_err, _fit_model_err, rtol=1e-2)
+
 
 class TestIndexedFitWithSimpleErrors(AbstractTestFit, unittest.TestCase):
 
@@ -391,7 +410,6 @@ class TestIndexedFitWithSimpleErrors(AbstractTestFit, unittest.TestCase):
         self.run_test_for_all_fits(
             dict(self._ref_prop_dict,
                 parameter_values=self._nominal_fit_result_pars,
-                poi_values=self._nominal_fit_result_pars,
                 model=self._nominal_fit_result_model,
                 did_fit=True,
                 cost_function_value=self._nominal_fit_result_cost,
@@ -531,7 +549,6 @@ class TestIndexedFitWithMatrixErrors(AbstractTestFit, unittest.TestCase):
         self.run_test_for_all_fits(
             dict(self._ref_prop_dict,
                 parameter_values=self._nominal_fit_result_pars,
-                poi_values=self._nominal_fit_result_pars,
                 model=self._nominal_fit_result_model,
                 did_fit=True,
                 cost_function_value=self._nominal_fit_result_cost,
